
#include <ntifs.h>
#include "preprocessor.h"
#include "utils-mem.h"


volatile ULONG __security_cookie = __LINE__;


PVOID HeapMemoryAlloc(POOL_TYPE PoolType, SIZE_T NumberOfBytes)
{
   PVOID ret = NULL;
   DEBUG_ENTER_FUNCTION("PoolType=%u; NumberOfBytes=%u", PoolType, NumberOfBytes);

   ret = ExAllocatePoolWithTag(PoolType, NumberOfBytes, PROJECT_POOL_TAG);

   DEBUG_EXIT_FUNCTION("0x%p", ret);
   return ret;
}


VOID HeapMemoryFree(PVOID Address)
{
   DEBUG_ENTER_FUNCTION("Address=0x%p", Address);

   ExFreePoolWithTag(Address, PROJECT_POOL_TAG);

   DEBUG_EXIT_FUNCTION_VOID();
   return;
}


PVOID HeapMemoryAllocPaged(SIZE_T NumberOfBytes)
{
   PVOID ret = NULL;
   DEBUG_ENTER_FUNCTION("NumberOfBytes=%u", NumberOfBytes);

   ret = HeapMemoryAlloc(PagedPool, NumberOfBytes);

   DEBUG_EXIT_FUNCTION("0x%p", ret);
   return ret;
}


PVOID HeapMemoryAllocNonPaged(SIZE_T NumberOfBytes)
{
   PVOID ret = NULL;
   DEBUG_ENTER_FUNCTION("NumberOfBytes=%u", NumberOfBytes);

   ret = HeapMemoryAlloc(NonPagedPool, NumberOfBytes);

   DEBUG_EXIT_FUNCTION("0x%p", ret);
   return ret;
}


NTSTATUS VirtualMemoryAllocUser(SIZE_T NumberOfBytes, ULONG Protection, PVOID *Address)
{
   PVOID tmp = NULL;
   NTSTATUS status = STATUS_UNSUCCESSFUL;
   DEBUG_ENTER_FUNCTION("NumberOfBytes=%u; Protection=0x%x; Address=0x%p", NumberOfBytes, Protection, Address);

   *Address = NULL;
   status = ZwAllocateVirtualMemory(NtCurrentProcess(), &tmp, 0, &NumberOfBytes, MEM_RESERVE | MEM_COMMIT, Protection);
   if (NT_SUCCESS(status))
      *Address = tmp;

   DEBUG_EXIT_FUNCTION("0x%x, *Address=0x%p", status, *Address);
   return status;
}


VOID VirtualMemoryFreeUser(PVOID Address)
{
   SIZE_T tmp = 0;
   NTSTATUS status = STATUS_UNSUCCESSFUL;
   DEBUG_ENTER_FUNCTION("Address=0x%p", Address);

   status = ZwFreeVirtualMemory(NtCurrentProcess(), &Address, &tmp, MEM_RELEASE);
   if (!NT_SUCCESS(status)) {
      DEBUG_ERROR("ZwFreeVirtualMemory failed at address 0x%p with status 0x%x", Address, status);
   }

   DEBUG_EXIT_FUNCTION_VOID()
   return;
}


NTSTATUS MemoryCopyToUser(PVOID Destionation, PVOID Source, SIZE_T NumberOfBytes)
{
   NTSTATUS status = STATUS_UNSUCCESSFUL;
   DEBUG_ENTER_FUNCTION("Destination=0x%p; Source=0x%p; NumberOfBytes=%u", Destionation, Source, NumberOfBytes);

   __try {
      ProbeForWrite(Destionation, NumberOfBytes, 1);
      RtlCopyMemory(Destionation, Source, NumberOfBytes);
      status = STATUS_SUCCESS;
   } __except (EXCEPTION_EXECUTE_HANDLER) {
      status = GetExceptionCode();
   }

   DEBUG_EXIT_FUNCTION("0x%x", status);
   return status;
}


NTSTATUS MemoryPointerCopyToUser(PVOID *Destination, PVOID Source)
{
   NTSTATUS status = STATUS_UNSUCCESSFUL;
   DEBUG_ENTER_FUNCTION("Destination=0x%p; Source=0x%p", Destination, Source);

   status = MemoryCopyToUser(Destination, &Source, sizeof(PVOID));

   DEBUG_EXIT_FUNCTION("0x%x", status);
   return status;
}


NTSTATUS MemoryCopyToKernel(PVOID Destionation, PVOID Source, SIZE_T NumberOfBytes)
{
   NTSTATUS status = STATUS_UNSUCCESSFUL;
   DEBUG_ENTER_FUNCTION("Destination=0x%p; Source=0x%p; NumberOfBytes=%u", Destionation, Source, NumberOfBytes);

   __try {
      ProbeForRead(Source, NumberOfBytes, 1);
      RtlCopyMemory(Destionation, Source, NumberOfBytes);
      status = STATUS_SUCCESS;
   } __except (EXCEPTION_EXECUTE_HANDLER) {
      status = GetExceptionCode();
   }

   DEBUG_EXIT_FUNCTION("0x%x", status);
   return status;
}


NTSTATUS MemoryPointerCopyToKernel(PVOID *Destination, PVOID *Source)
{
   NTSTATUS status = STATUS_UNSUCCESSFUL;
   DEBUG_ENTER_FUNCTION("Destination=0x%p; Source=0x%p", Destination, Source);

   status = MemoryCopyToKernel(Destination, Source, sizeof(PVOID));

   DEBUG_EXIT_FUNCTION("0x%x", status);
   return status;
}


NTSTATUS MemoryCopyInUser(PVOID Destination, PVOID Source, SIZE_T NumberOfBytes)
{
   NTSTATUS status = STATUS_UNSUCCESSFUL;
   DEBUG_ENTER_FUNCTION("Destination=0x%p; Source=0x%p; NumberOfBytes=%u", Destination, Source, NumberOfBytes);

   __try {
      ProbeForRead(Source, NumberOfBytes, 1);
      ProbeForWrite(Destination, NumberOfBytes, 1);
      RtlCopyMemory(Destination, Source, NumberOfBytes);
      status = STATUS_SUCCESS;
   } __except (EXCEPTION_EXECUTE_HANDLER) {
      status = GetExceptionCode();
   }

   DEBUG_EXIT_FUNCTION("0x%x", status);
   return status;
}


NTSTATUS MemoryCopySafe(PVOID Destination, PVOID Source, SIZE_T NumberOfBytes)
{
   NTSTATUS status = STATUS_UNSUCCESSFUL;
   DEBUG_ENTER_FUNCTION("Destination=0x%p; Source=0x%p; NumberOfbytes=%u", Destination, Source, NumberOfBytes);

   if (Destination > MmHighestUserAddress && Source > MmHighestUserAddress) {
      RtlCopyMemory(Destination, Source, NumberOfBytes);
      status = STATUS_SUCCESS;
   } else if (Destination > MmHighestUserAddress) {
      status = MemoryCopyToKernel(Destination, Source, NumberOfBytes);
   } else if (Source > MmHighestUserAddress) {
      status = MemoryCopyToUser(Destination, Source, NumberOfBytes);
   } else status = MemoryCopyInUser(Destination, Source, NumberOfBytes);

   DEBUG_EXIT_FUNCTION("0x%x", status);
   return status;
}


NTSTATUS MemoryPointerCopySafe(PVOID *Destination, PVOID *Source)
{
   NTSTATUS status = STATUS_UNSUCCESSFUL;
   DEBUG_ENTER_FUNCTION("Destination=0x%p; Source=0x%p", Destination, Source);

   status = MemoryCopySafe(Destination, Source, sizeof(PVOID));

   DEBUG_EXIT_FUNCTION("0x%x", status);
   return status;
}

NTSTATUS MemoryPointerCopyInUser(PVOID *Destination, PVOID *Source)
{
   NTSTATUS status = STATUS_UNSUCCESSFUL;
   DEBUG_ENTER_FUNCTION("Destination=0x%p; Source=0x%p", Destination, Source);

//   __security_cookie = RtlRandom((PULONG)&__security_cookie);
   status = MemoryCopyInUser(Destination, Source, sizeof(PVOID));

   DEBUG_EXIT_FUNCTION("0x%x", status);
   return status;
}
