
#include <ntifs.h>
#include "preprocessor.h"
#include "utils-mem.h"
#include "utils-stack.h"
#include "utils-stacked-memory.h"


/************************************************************************/
/*                              HELPER TOUTINES                         */
/************************************************************************/

static PUTILS_STACKED_MEMORY_RECORD _StackedMemoryAllocRecord(PUTILS_STACK StackedMemory, EStackedMemoryAllocType AllocType, PVOID Address)
{
   PUTILS_STACKED_MEMORY_RECORD ret = NULL;
   DEBUG_ENTER_FUNCTION("StackedMemory=0x%p; AllocType=%u; Address=0x%p", StackedMemory, AllocType, Address);

   ret = (PUTILS_STACKED_MEMORY_RECORD)HeapMemoryAlloc(StackedMemory->PoolType, sizeof(UTILS_STACKED_MEMORY_RECORD));
   if (ret != NULL) {
      ret->Address = Address;
      ret->AllocType = AllocType;
   }

   DEBUG_EXIT_FUNCTION("0x%p", ret);
   return ret;
}


/************************************************************************/
/*                        PUBLIC ROUTINES                               */
/************************************************************************/


NTSTATUS StackedMemoryCreate(POOL_TYPE PoolType, PUTILS_STACK *StackedMemory)
{
   PUTILS_STACK tmpStackedMemory = NULL;
   NTSTATUS status = STATUS_UNSUCCESSFUL;
   DEBUG_ENTER_FUNCTION("PoolType=%u; StackedMemory=0x%p", PoolType, StackedMemory);

   *StackedMemory = NULL;
   status = StackCreate(PoolType, &tmpStackedMemory);
   if (NT_SUCCESS(status))
      *StackedMemory = tmpStackedMemory;

   DEBUG_EXIT_FUNCTION("0x%x, *StackedMemory=0x%p", status, *StackedMemory);
   return status;
}



NTSTATUS StackedMemoryHeapAlloc(PUTILS_STACK StackedMemory, POOL_TYPE PoolType, SIZE_T NumberOfBytes, PVOID *Address)
{
   PVOID tmpAddress = NULL;
   PUTILS_STACKED_MEMORY_RECORD stackRecord = NULL;
   NTSTATUS status = STATUS_UNSUCCESSFUL;
   DEBUG_ENTER_FUNCTION("StackedMemory=0x%p; PoolType=%u; NumberOfBytes=%u; Address=0x%p", StackedMemory, PoolType, NumberOfBytes, Address);

   *Address = NULL;
   tmpAddress = HeapMemoryAlloc(PoolType, NumberOfBytes);
   if (tmpAddress != NULL) {
      stackRecord = _StackedMemoryAllocRecord(StackedMemory, smatHeap, tmpAddress);
      if (stackRecord != NULL) {
         StackPushNoAlloc(StackedMemory, &stackRecord->StackItem);
         *Address = tmpAddress;
         status = STATUS_SUCCESS;
      } else status = STATUS_INSUFFICIENT_RESOURCES;

      if (!NT_SUCCESS(status))
         HeapMemoryFree(tmpAddress);
   } else status = STATUS_INSUFFICIENT_RESOURCES;

   DEBUG_EXIT_FUNCTION("0x%x, *Address=0x%p", *Address);
   return status;
}


NTSTATUS StackedMemoryHeapAllocPaged(PUTILS_STACK StackedMemory, SIZE_T NumberOfBytes, PVOID *Address)
{
   NTSTATUS status = STATUS_UNSUCCESSFUL;
   DEBUG_ENTER_FUNCTION("StackedMemory=0x%p; NumberOfBytes=%u; Address=0x%p", StackedMemory, NumberOfBytes, Address);

   status = StackedMemoryHeapAlloc(StackedMemory, PagedPool, NumberOfBytes, Address);

   DEBUG_EXIT_FUNCTION("0x%x, *Address=0x%p", *Address);
   return status;
}


NTSTATUS StackedMemoryHeapAllocNonPaged(PUTILS_STACK StackedMemory, SIZE_T NumberOfBytes, PVOID *Address)
{
   NTSTATUS status = STATUS_UNSUCCESSFUL;
   DEBUG_ENTER_FUNCTION("StackedMemory=0x%p; NumberOfBytes=%u; Address=0x%p", StackedMemory, NumberOfBytes, Address);

   status = StackedMemoryHeapAlloc(StackedMemory, NonPagedPool, NumberOfBytes, Address);

   DEBUG_EXIT_FUNCTION("0x%x, *Address=0x%p", *Address);
   return status;
}


NTSTATUS StackedMemoryVirtualAllocUser(PUTILS_STACK StackedMemory, SIZE_T NumberOfBytes, ULONG Protection, PVOID *Address)
{
   PVOID tmpAddress = NULL;
   NTSTATUS status = STATUS_UNSUCCESSFUL;
   PUTILS_STACKED_MEMORY_RECORD stackRecord = NULL;
   DEBUG_ENTER_FUNCTION("StackedMemory=0x%p; NumberOfBytes=%u; Protection=0x%x; Address=0x%p", StackedMemory, NumberOfBytes, Protection, Address);

   status = VirtualMemoryAllocUser(NumberOfBytes, Protection, &tmpAddress);
   if (NT_SUCCESS(status)) {
      stackRecord = _StackedMemoryAllocRecord(StackedMemory, smatVirtual, tmpAddress);
      if (stackRecord != NULL) {
         StackPushNoAlloc(StackedMemory, &stackRecord->StackItem);
         status = STATUS_SUCCESS;
         *Address = tmpAddress;
      } else status = STATUS_INSUFFICIENT_RESOURCES;

      if (!NT_SUCCESS(status))
         VirtualMemoryFreeUser(tmpAddress);
   }

   DEBUG_EXIT_FUNCTION("0x%x, *Address=0x%p", status, *Address);
   return status;
}


VOID StackedMemoryLastFree(PUTILS_STACK StackedMemory)
{
   PUTILS_STACK_ITEM stackItem = NULL;
   PUTILS_STACKED_MEMORY_RECORD stackRecord = NULL;
   DEBUG_ENTER_FUNCTION("StackedMemory=0x%p", StackedMemory);

   if (!StackEmpty(StackedMemory)) {
      stackItem = StackPopNoFree(StackedMemory);
      stackRecord = CONTAINING_RECORD(stackItem, UTILS_STACKED_MEMORY_RECORD, StackItem);
      switch (stackRecord->AllocType) {
      case smatHeap:
         HeapMemoryFree(stackRecord->Address);
         break;
      case smatVirtual:
         VirtualMemoryFreeUser(stackRecord->Address);
         break;
      default:
         DEBUG_ERROR("Invalid stacked memory record type (%u)", stackRecord->AllocType);
         KeBugCheck(0);
         break;
      }

      HeapMemoryFree(stackRecord);
   }

   DEBUG_EXIT_FUNCTION_VOID();
   return;
}


VOID StackedMemoryAllFree(PUTILS_STACK StackedMemory)
{
   PUTILS_STACK_ITEM stackItem = NULL;
   PUTILS_STACKED_MEMORY_RECORD stackRecord = NULL;
   DEBUG_ENTER_FUNCTION("StackedMemory=0x%p", StackedMemory);

   while (!StackEmpty(StackedMemory)) {
      stackItem = StackPopNoFree(StackedMemory);
      stackRecord = CONTAINING_RECORD(stackItem, UTILS_STACKED_MEMORY_RECORD, StackItem);
      switch (stackRecord->AllocType) {
      case smatHeap:
         HeapMemoryFree(stackRecord->Address);
         break;
      case smatVirtual:
         VirtualMemoryFreeUser(stackRecord->Address);
         break;
      default:
         DEBUG_ERROR("Invalid stacked memory record type (%u)", stackRecord->AllocType);
         KeBugCheck(0);
         break;
      }

      HeapMemoryFree(stackRecord);
   }

   DEBUG_EXIT_FUNCTION_VOID();
   return;
}


VOID StackedMemoryDestroy(PUTILS_STACK StackedMemory)
{
   PUTILS_STACK_ITEM stackItem = NULL;
   PUTILS_STACKED_MEMORY_RECORD stackRecord = NULL;
   DEBUG_ENTER_FUNCTION("StackedMemory=0x%p", StackedMemory);

   while (!StackEmpty(StackedMemory)) {
      stackItem = StackPopNoFree(StackedMemory);
      stackRecord = CONTAINING_RECORD(stackItem, UTILS_STACKED_MEMORY_RECORD, StackItem);
      HeapMemoryFree(stackRecord);
   }

   StackDestroy(StackedMemory);

   DEBUG_EXIT_FUNCTION_VOID();
   return;
}
